package ij.plugin.filter;

import ij.*;
import ij.gui.*;
import ij.plugin.*;
import ij.process.*;
import ij.measure.*;
import ij.util.Tools;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.beans.*;
import ij.plugin.filter.PlugInFilter;

import javax.swing.JComponent;
import javax.swing.JSlider;
import javax.swing.JLabel;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.text.*;
import javax.swing.event.*;
import javax.swing.JFormattedTextField;
import javax.swing.JSlider;
import javax.swing.JSpinner;
import javax.swing.border.BevelBorder;
import javax.swing.border.Border;
import javax.swing.BorderFactory;

/** This plugin implements the Edit/Scale command. */
//public class Scaler implements PlugInFilter, DocumentListener {

/** This plugin implements the Edit/Scale command. */
public class Scaler implements PlugIn, DocumentListener, FocusListener {
    private ImagePlus imp;
    private static String xstr = "0.5";
    private static String ystr = "0.5";
    private static int newWidth, newHeight;
    private static boolean newWindow = true;
    private static boolean interpolate = true;
    private static boolean fillWithBackground;
    private static boolean processStack = true;
    private static double xscale=0.5;
    private static double yscale=0.5;
    private String title = "Untitled";
    private Vector<JComponent> fields;
    private double bgValue;
    private boolean constainAspectRatio = true;
    private JTextField widthField, heightField;
    private Document[] doc= new Document[4];
    private Rectangle r;
    private Object fieldWithFocus;
    private JFormattedTextField xtextField, ytextField;
    private JSpinner xspinner, yspinner;
    private Vector<JComponent> spinnerFields=new Vector<JComponent>(), spinners=new Vector<JComponent>();
    private BevelBorder bb ;//new BevelBorder for LOWER or RAISED type
    private Border grayBorder = BorderFactory.createBevelBorder(bb.RAISED, Color.gray, Color.lightGray);

        public void run(String arg) {
                imp = IJ.getImage();
                Roi roi = imp.getRoi();
                if (roi!=null && !roi.isArea())
                        imp.killRoi(); // ignore any line selection
                ImageProcessor ip = imp.getProcessor();
                if (!showDialog(ip))
                        return;
                if (ip.getWidth()>1 && ip.getHeight()>1)
                        ip.setInterpolate(interpolate);
                else
                        ip.setInterpolate(false);
                ip.setBackgroundValue(bgValue);
                imp.startTiming();
                try {
                        if (newWindow && imp.getStackSize()>1 && processStack)
                                createNewStack(imp, ip);
                        else
                                scale(ip);
                }
                catch(OutOfMemoryError o) {
                        IJ.outOfMemory("Scale");
                }
                IJ.showProgress(1.0);
        }

        void createNewStack(ImagePlus imp, ImageProcessor ip) {
                Rectangle r = ip.getRoi();
                boolean crop = r.width!=imp.getWidth() || r.height!=imp.getHeight();
                int nSlices = imp.getStackSize();
            ImageStack stack1 = imp.getStack();
            ImageStack stack2 = new ImageStack(newWidth, newHeight);
                ImageProcessor ip1, ip2;
                boolean interp = interpolate;
                if (imp.getWidth()==1 || imp.getHeight()==1)
                        interp = false;
                for (int i=1; i<=nSlices; i++) {
                        IJ.showStatus("Scale: " + i + "/" + nSlices);
                        ip1 = stack1.getProcessor(i);
                        String label = stack1.getSliceLabel(i);
                        if (crop) {
                                ip1.setRoi(r);
                                ip1 = ip1.crop();
                        }
                        ip1.setInterpolate(interp);
                        ip2 = ip1.resize(newWidth, newHeight);
                        if (ip2!=null)
                                stack2.addSlice(label, ip2);
                        IJ.showProgress(i, nSlices);
                }
                ImagePlus imp2 = imp.createImagePlus();
                imp2.setStack(title, stack2);
                Calibration cal = imp2.getCalibration();
                if (cal.scaled()) {
                        cal.pixelWidth *= 1.0/xscale;
                        cal.pixelHeight *= 1.0/yscale;
                }
                IJ.showProgress(1.0);
                imp2.show();
                imp2.changes = true;
        }

        void scale(ImageProcessor ip) {
                if (newWindow) {
                        Rectangle r = ip.getRoi();
                        ImagePlus imp2 = imp.createImagePlus();
                        imp2.setProcessor(title, ip.resize(newWidth, newHeight));
                        Calibration cal = imp2.getCalibration();
                        if (cal.scaled()) {
                                cal.pixelWidth *= 1.0/xscale;
                                cal.pixelHeight *= 1.0/yscale;
                        }
                        imp2.show();
                        imp.trimProcessor();
                        imp2.trimProcessor();
                        imp2.changes = true;
                } else {
                        if (processStack && imp.getStackSize()>1) {
                                Undo.reset();
                                StackProcessor sp = new StackProcessor(imp.getStack(), ip);
                                sp.scale(xscale, yscale, bgValue);
                        } else {
                                ip.snapshot();
                                Undo.setup(Undo.FILTER, imp);
                                ip.scale(xscale, yscale);
                        }
                        imp.killRoi();
                        imp.updateAndDraw();
                        imp.changes = true;
                }
        }

        boolean showDialog(ImageProcessor ip) {
                int bitDepth = imp.getBitDepth();
                boolean isStack = imp.getStackSize()>1;
        r = ip.getRoi();
        int width = newWidth;
                if (width==0) width = r.width;
        int height = (int)((double)width*r.height/r.width);
        //xscale = Tools.parseDouble(xstr, 0.0);
        //yscale = Tools.parseDouble(ystr, 0.0);
        if (xscale!=0.0 && yscale!=0.0) {
                width = (int)(r.width*xscale);
                height = (int)(r.height*yscale);
        } else {
                xstr = "-";
                ystr = "-";
        }
                GenericDialog gd = new GenericDialog("Scale");
                String scaleXLabel="X Scale(0.05-25):";
                String scaleYLabel="Y Scale(0.05-25):";
                double min=0.05, max=25.0, value=0.5;
                int swidth=80, sheight=21;
                if(xscale < min || yscale < min){xscale=min;yscale=min;}
                //boolean paintticks=true, paintLabels=true;
                JLabel transformLabel = new JLabel(" ");
                //transformLabel.setBorder(grayBorder);
                gd.addPanel(transformLabel, GridBagConstraints.WEST, new Insets(0, 0, 10, 10));
                gd.addSpinner(scaleXLabel, min, max, xscale, swidth, sheight);
                gd.addSpinner(scaleYLabel, min, max, yscale, swidth, sheight);
                spinnerFields = gd.getSpinnerFields();
                int spinnerSize = spinnerFields.size();
                JFormattedTextField[] formatTextField = new JFormattedTextField[spinnerSize];
                spinners = gd.getSpinners();
                for (int i=0; i<spinnerSize; i++) {
                    formatTextField[i]=(JFormattedTextField)spinnerFields.elementAt(i);
                    formatTextField[i].addFocusListener(this);
                    doc[i]=formatTextField[i].getDocument();
                    doc[i].addDocumentListener(this);
                        //((JFormattedTextField)spinnerFields.elementAt(i)).getDocument().addDocumentListener(this);
                        //doc[i]=((JFormattedTextField)spinnerFields.elementAt(i)).getDocument();
                        //((JFormattedTextField)spinnerFields.elementAt(i)).addFocusListener(this);
                        //((JSpinner)spinners.elementAt(i)).addChangeListener(this);
                }
                xtextField = formatTextField[0];//(JFormattedTextField)spinnerFields.elementAt(0);
                ytextField = formatTextField[1];// (JFormattedTextField)spinnerFields.elementAt(1);
                xspinner = (JSpinner)spinners.elementAt(0);
                yspinner = (JSpinner)spinners.elementAt(1);
                gd.setInsets(5, 0, 5);
                gd.addStringField("Width (pixels): ", ""+width);
                gd.addStringField("Height (pixels): ", ""+height);
                fields = gd.getStringFields();
                int fieldSize = fields.size();
                JTextField[] textField = new JTextField[fieldSize];
                for (int i=0; i<fieldSize; i++) {
                    textField[i]=(JTextField)fields.elementAt(i);
                    textField[i].addFocusListener(this);
                    doc[i+spinnerSize]=textField[i].getDocument();
                    doc[i+spinnerSize].addDocumentListener(this);
                    formatTextField[i].setColumns(textField[i].getColumns()-1);
                        //((JTextField)fields.elementAt(i)).getDocument().addDocumentListener(this);
                        //doc[i+spinnerSize]=((JTextField)fields.elementAt(i)).getDocument();
                        //((JTextField)fields.elementAt(i)).addFocusListener(this);
                }
                widthField = textField[0];//(JTextField)fields.elementAt(0);
                heightField = textField[1];//(JTextField)fields.elementAt(1);
        fieldWithFocus = xtextField;
                gd.addCheckbox("Interpolate", interpolate);
                if (bitDepth==8 || bitDepth==24)
                        gd.addCheckbox("Fill with Background Color", fillWithBackground);
                if (isStack)
                        gd.addCheckbox("Process Entire Stack", processStack);
                gd.addCheckbox("Create New Window", newWindow);
                title = WindowManager.getUniqueName(imp.getTitle());
                gd.setInsets(10, 0, 0);
                gd.addStringField("Title:", title, 12);
                gd.showDialog();
                if (gd.wasCanceled())
                        return false;
                    Double dx = (Double)xspinner.getValue();
                    Double dy = (Double)yspinner.getValue();
                xscale = dx.doubleValue();// gd.getNextNumbers();
                yscale = xscale;//dy.doubleValue();//  gd.getNextNumbers();
                String wstr = gd.getNextString();
                newWidth = (int)Tools.parseDouble(wstr, 0);
                String hstr = gd.getNextString();
                newHeight = (int)Tools.parseDouble(hstr, 0);
                if (newHeight!=0 && (wstr.equals("-") || wstr.equals("0")))
                newWidth= (int)(newHeight*(double)r.width/r.height);
                if (newWidth==0 || newHeight==0) {
                        IJ.error("Invalid width or height entered");
                        return false;
                }
                if (xscale>25.0) xscale = 25.0;
                if (yscale>25.0) yscale = 25.0;
                if (xscale>0.0 && yscale>0.0) {
                        newWidth = (int)(r.width*xscale);
                        newHeight = (int)(r.height*yscale);
                }
                interpolate = gd.getNextBoolean();
                if (bitDepth==8 || bitDepth==24)
                        fillWithBackground = gd.getNextBoolean();
                if (isStack)
                        processStack = gd.getNextBoolean();
                newWindow = gd.getNextBoolean();
                if (!newWindow && xscale==0.0) {
                        xscale = (double)newWidth/r.width;
                        yscale = (double)newHeight/r.height;
                }
                title = gd.getNextString();

                if (fillWithBackground) {
                        Color bgc = Toolbar.getBackgroundColor();
                        if (bitDepth==8)
                                bgValue = ip.getBestIndex(bgc);
                        else if (bitDepth==24)
                                bgValue = bgc.getRGB();
                } else {
                        if (bitDepth==8)
                                bgValue = ip.isInvertedLut()?0.0:255.0; // white
                        else if (bitDepth==24)
                                bgValue = 0xffffffff; // white
                }
                return true;
        }

      public void textFieldUpdate(DocumentEvent e) {
          Document document = (Document)e.getDocument();
          double newXScale = xscale;
          double newYScale = yscale;
          Object source=null;
        for(int i=0; i< doc.length; i++)
            if(document.equals(doc[i])){
                switch (i) {
                case 0:
                    source = xtextField;
                    break;
                case 1:
                    source = ytextField;
                    break;
                case 2:
                    source = widthField;
                    break;
                case 3:
                    source = heightField;
                    break;
                }
            }
        if (source==xtextField && fieldWithFocus==xtextField) {
            String newXText = xtextField.getText();
            newXScale = Tools.parseDouble(newXText,0.0);
            if (newXScale==0) return;
            if (newXScale!=xscale) {
                int newWidth = (int)(newXScale*r.width);
                widthField.setText(""+newWidth);
                if (constainAspectRatio) {
                    ytextField.setText(newXText);
                    int newHeight = (int)(newXScale*r.height);
                    heightField.setText(""+newHeight);
                }
            }
        } else if (source==ytextField && fieldWithFocus==ytextField) {
            String newYText = ytextField.getText();
            newYScale = Tools.parseDouble(newYText,0.0);
            if (newYScale==0) return;
            if (newYScale!=yscale) {
                int newHeight = (int)(newYScale*r.height);
                heightField.setText(""+newHeight);
                /*if (constainAspectRatio) {
                    xtextField.setText(newYText);
                    int newWidth = (int)(newYScale*r.width);
                    heightField.setText(""+newWidth);
                }*/
            }
        } else if (source==widthField && fieldWithFocus==widthField) {
            int newWidth = (int)Tools.parseDouble(widthField.getText(), 0.0);
            if (newWidth!=0) {
                int newHeight = (int)(newWidth*(double)r.height/r.width);
                heightField.setText(""+newHeight);
                xtextField.setText("-");
                ytextField.setText("-");
                newXScale = 0.0;
                newYScale = 0.0;
            }
        }
                xscale = newXScale;
                yscale = newYScale;
        }

        public void insertUpdate(DocumentEvent e){
            textFieldUpdate(e);
        }
        public void removeUpdate(DocumentEvent e){
            textFieldUpdate(e);
        }
        public void changedUpdate(DocumentEvent e){
        }
        public void focusGained(FocusEvent e) {
        fieldWithFocus = e.getSource();
                if (fieldWithFocus==widthField)
            constainAspectRatio = true;
        else if (fieldWithFocus==ytextField)
            constainAspectRatio = false;
        }

        public void focusLost(FocusEvent e) {
            constainAspectRatio = true;
        }

}
